#include <std.h>
#include <stdio.h>
#include <func/apt.h>
#include <func/config.h>
#include <func/messages.h>
#include <func/dependency.h>
#include "pkgmgrctrl.h"

#include <pthread.h>
//old #include <sys/types.h>
#include <sys/socket.h>
#include <sys/vfs.h>

#include "acquire.h"

pkgAcquireBackend::pkgAcquireBackend(EDeityUI *_Master) : pkgAcquire(this), Master(_Master)
{
	ASSERT(Master != 0);
	signalfd = -1;
	NoCacheInit = false;
}

pkgAcquireBackend::~pkgAcquireBackend()
{
	if (signalfd > 0) 
	{
		NoCacheInit = !NoCacheInit;
		MessageSend(signalfd, MsgId_Acquire_Shutdown, &NoCacheInit, sizeof(bool));
		close(signalfd);
	}
}

int pkgAcquireBackend::AcquireUpdate()
{
	int fds[2];

	if (socketpair(PF_UNIX, SOCK_STREAM, 0, fds) == -1)
		return -1;

	signalfd = fds[0];
	Master->CreateAcquireUI();
	Master->AcqItf->acqState = EDeityUI::AcquireUI::UpdateRun;

	pthread_t thread;
	pthread_create(&thread, 0, DoAcquireUpdate, this);
	pthread_detach(thread);

	return fds[1];
}

void *pkgAcquireBackend::DoAcquireUpdate(void *data)
{
	pkgAcquireBackend *This = (pkgAcquireBackend *)data;
	This->_AcquireUpdate();
	delete This;
	pthread_exit(0);
	return 0;
}

void pkgAcquireBackend::_AcquireUpdate()
{
	// Get the source list
	pkgSourceList List;
	if (List.ReadMainList() == false)
	{
		TextMessage(signalfd, MsgId_Acquire_Fatal, _("Cleaning out lists directories FAILED. Please check your permissions."));
		return;
	}

	// Lock the list directory
	FileFd Lock;
	if (_config->FindB("Debug::NoLocking", false) == false)
		Lock.Fd(GetLock(_config->FindDir("Dir::State::Lists") + "lock"));
			if (_error->PendingError() == true)
				_error->Discard();


	// Populate it with the source selection
	if (List.GetIndexes(this) == false)
		return;

	switch (Run())
	{
		case Failed:
		case Cancelled:
			AckErrors();
			NoCacheInit = true;
			return;

		case Continue:
			break;
	}
	
	Lock.Close();

	// Clean out any old list files
	if (Clean(_config->FindDir("Dir::State::lists")) == false ||
	    Clean(_config->FindDir("Dir::State::lists") + "partial/") == false)
	{
		TextMessage(signalfd, MsgId_Acquire_Fatal, _("Cleaning out lists directories FAILED. Please check your permissions."));
		return;
	}
}

int pkgAcquireBackend::AcquireUpgrade()
{
	// Do the upgrade

	if (pkgAllUpgrade(*Cache) == false)
	{
		// Bad things happened with the upgrade, tell the
		// user about it and then return to the coreui.
		capt_show_broken(*Cache);
		return -1;
	}

	return AcquireRun();  // Complete the stuff
}

int pkgAcquireBackend::AcquireRun()
{
	int fds[2];

	if (socketpair(PF_UNIX, SOCK_STREAM, 0, fds) == -1)
		return -1;

	signalfd = fds[0];
	Master->CreateAcquireUI();
	Master->AcqItf->acqState = EDeityUI::AcquireUI::PackageRun;
	
	Start();

	pthread_t thread;
	pthread_create(&thread, 0, DoAcquireRun, this);
	pthread_detach(thread);

/*	if (fork() == 0)
		DoAcquireRun(this);
*/

	return fds[1];
}

void *pkgAcquireBackend::DoAcquireRun(void *data)
{
	pkgAcquireBackend *This = (pkgAcquireBackend *)data;
	This->_AcquireRun();
	delete This;
	pthread_exit(0);
	return 0;
}

void pkgAcquireBackend::_AcquireRun()
{
	// Hopefully we can't get called in vain

	// Lock the archive directory
	FileFd Lock;
	if (_config->FindB("Debug::NoLocking", false) == false)
		Lock.Fd(GetLock(_config->FindDir("Dir::Cache::Archives") + "lock"));
			if (_error->PendingError() == true)
				_error->Discard();

	if (_error->PendingError() == true)
	{
		string s;
		_error->PopMessage(s);
		TextMessage(signalfd, MsgId_Acquire_Fatal, s.c_str());
		NoCacheInit = true;
		return;
	}

	// Read the source list
	pkgSourceList List;
	if (List.ReadMainList() == false)
	{
		TextMessage(signalfd, MsgId_Acquire_Fatal, _("The list of sources could not be read."));
		return;
	}

	// Create the package manager and prepare to download

	//old	pkgDPkgPM PM(*Cache);
	pkgMgrCtrl PM(signalfd, *Cache);
	if (PM.GetArchives(this, &List, &Cache->GetPkgRecs()) == false || _error->PendingError() == true)
	{
		TextMessage(signalfd, MsgId_Acquire_Fatal, _("Error creating package manager"));
		return;
	}

	double FetchBytes = FetchNeeded();
	double FetchPBytes = PartialPresent();
	double DebBytes = TotalNeeded();

	// Only boot up the interface if there is something to download
	if (FetchBytes > 0)
	{
		if (DebBytes != (*Cache)->DebSize())
		{
			TextMessage(signalfd, MsgId_Acquire_Fatal, _("(%.0f, %.0f) How odd.. The sizes didn't match, email apt@packages.debian.org"), DebBytes, (*Cache)->DebSize());
			return;
		}

		// Check for enough free space
		struct statfs statbuf;
		string OutputDir = _config->FindDir("Dir::Cache::Archives");

		if (statfs(OutputDir.c_str(), &statbuf) != 0)
			TextMessage(signalfd, MsgId_Acquire_Warning, _("Couldn't determine free space in %s"), OutputDir.c_str());

		if (unsigned (statbuf.f_bfree) < (FetchBytes - FetchPBytes) / statbuf.f_bsize)
		{
			TextMessage(signalfd, MsgId_Acquire_Fatal, _("Sorry, you don't have enough free space in %s"), OutputDir.c_str());
			return;
		}

		// Tell them what we want  FIXME
		if (DebBytes != FetchBytes)
			TextMessage(signalfd, MsgId_Acquire_Info, _("%sB of %sB needs to be downloaded."), SizeToStr(FetchBytes).c_str(), SizeToStr(DebBytes).c_str());
		else
			TextMessage(signalfd, MsgId_Acquire_Info, _("%sB needs to be downloaded."), SizeToStr(DebBytes).c_str());

		// Size delta
		if ((*Cache)->UsrSize() >= 0)
			TextMessage(signalfd, MsgId_Acquire_Info, _("Afterwards, %sB will be used."), SizeToStr((*Cache)->UsrSize()).c_str());
		else
			TextMessage(signalfd, MsgId_Acquire_Info, _("Afterwards, %sB will be freed."), SizeToStr(-1 * (*Cache)->UsrSize()).c_str());
		if (_error->PendingError() == true)
		{
			AckErrors();
			return;
		}

		// Run it
		switch (Run())
		{
			case (Failed):
			case (Cancelled):
				Lock.Close();
				AckErrors();
				NoCacheInit = true;
				return;

			case (Continue):
				break;
		}
	
		// Print out errors
		bool Failed = false;
		bool Transient = false;

		for (ItemIterator It = ItemsBegin(); It != ItemsEnd(); It++)
		{
			Item *I = (*It);

			if (I->Status == Item::StatDone && I->Complete == true)
				continue;

			if (I->Status == Item::StatIdle)
			{
				Transient = true;
				Failed = true;
				continue;
			}
	
			// FIXME
			TextMessage(signalfd, MsgId_Acquire_Info, _("Failed to fetch %s: \n"), I->DescURI().c_str(), I->ErrorText.c_str());
	
			Failed = true;
		}
	}

	// Try to deal with missing package files
	if (PM.FixMissing() == false)
	{
		TextMessage(signalfd, MsgId_Acquire_Fatal, _("I was unable to correct the missing package files."));
		return;
	}

	// Unlock before installing packages
	_system->UnLock();

	// See if the UI is ready for us to run the package manager
//old	bool ready = false;
//old	unsigned long len = sizeof(bool);
//old	MsgId msgid;	
//old	do {
//old		MessageSend(signalfd, MsgId_PkgMgr_Start, 0, 0);
//old		MessageReceive(signalfd, msgid, &ready, len);
//old		ASSERT(msgid == MsgId_PkgMgr_Ready);
//old		ASSERT(len == sizeof(bool));
//old		if (!ready) sleep(3); // give things some time to happen
//old	} while (!ready);
	
	// Run dpkg
	//old	pkgMgrCtrl pmctrl(signalfd, PM); pmctrl.run();


	MessageSend(signalfd, MsgId_PkgMgr_Start, 0, 0);

	switch (PM.DoInstall())
	{
		case pkgPackageManager::Completed:
			TextMessage(signalfd, MsgId_PkgMgr_Warning, _("Package manager (dpkg) completed"), 0);
		break;
		
		case pkgPackageManager::Incomplete:
			TextMessage(signalfd, MsgId_PkgMgr_Warning, _("Package manager (dpkg) incomplete"), 0);
		break;
		
		case pkgPackageManager::Failed:
			TextMessage(signalfd, MsgId_PkgMgr_Warning, _("Package manager (dpkg) failed"), 0);
		break;
	}

	MessageSend(signalfd, MsgId_PkgMgr_Stop, 0, 0);

	return;
	// and exit out of thread
}

void pkgAcquireBackend::AckErrors()
{
	string s, q;

	while (_error->PendingError() == true)
	{	
		_error->PopMessage(q);
		s += " * ";
		s += q;
		s += "\n";
	}
	
	if (!s.empty())
		TextMessage(signalfd, MsgId_Error, s.c_str(), s.length());
}

void pkgAcquireBackend::Start()
{
	MorePulses = true;
	pkgAcquireStatus::Start();
	ID = 0;
	MessageSend(signalfd, MsgId_Acquire_Start, NULL, 0);
}

void pkgAcquireBackend::Stop()
{
	pkgAcquireStatus::Stop();
	MessageSend(signalfd, MsgId_Acquire_Stop, NULL, 0);
}

void pkgAcquireBackend::IMSHit(pkgAcquire::ItemDesc & Itm)
{
	Update = true;

	//if (Itm.Owner->Complete == true)
	//	return;

	struct MsgItemDesc Data;

	Data.FileSize = Itm.Owner->FileSize;
	Data.ID = Itm.Owner->ID;
	strncpy(Data.URI,         Itm.URI.c_str(), 200);
	strncpy(Data.Description, Itm.Description.c_str(), 200);
	strncpy(Data.ShortDesc,   Itm.ShortDesc.c_str(), 200);
	
	MessageSend(signalfd, MsgId_Site_Hit, &Data, sizeof(Data));
};

void pkgAcquireBackend::Fetch(pkgAcquire::ItemDesc & Itm)
{
	Itm.Owner->ID = ++ID;

	struct MsgItemDesc Data;

	Data.FileSize = Itm.Owner->FileSize;
	Data.ID = Itm.Owner->ID;
	strncpy(Data.URI,         Itm.URI.c_str(), 200);
	strncpy(Data.Description, Itm.Description.c_str(), 200);
	strncpy(Data.ShortDesc,   Itm.ShortDesc.c_str(), 200);
	
	MessageSend(signalfd, MsgId_Download_Start, &Data, sizeof(Data));

	Update = true;
};

void pkgAcquireBackend::Fail(pkgAcquire::ItemDesc & Itm)
{
	struct MsgItemFailed Data;

	Data.Itm.FileSize = Itm.Owner->FileSize;
	//Data.ID = Itm.Owner->ID;
	strncpy(Data.Itm.URI,         Itm.URI.c_str(), 200);
	strncpy(Data.Itm.Description, Itm.Description.c_str(), 200);
	strncpy(Data.Itm.ShortDesc,   Itm.ShortDesc.c_str(), 200);
	strncpy(Data.ErrorText,       Itm.Owner->ErrorText.c_str(), 200);
	
	MessageSend(signalfd, MsgId_Download_Failed, &Data, sizeof(Data));

	Update = true;
};

bool pkgAcquireBackend::Pulse(pkgAcquire * Owner)
{
	pkgAcquireStatus::Pulse(Owner);

	unsigned int J = 0;
	struct MsgDownloadItem Itm;

	for (Worker * I = Owner->WorkersBegin(); I != 0; I = Owner->WorkerStep(I), J++)
	{
		if ((I->CurrentItem == 0) && I->Status.empty())
			continue;

		if (I->CurrentItem == 0)
		{
			Itm.ID = 0;
			strncpy(Itm.Status, I->Status.c_str(), 100);
			*Itm.ShortDesc = 0;
		}
		else
		{
			Itm.ID = I->CurrentItem->Owner->ID;
			*Itm.Status = 0;
			strncpy(Itm.ShortDesc, I->CurrentItem->ShortDesc.c_str(), 100);
		}

		Itm.FileSize = I->TotalSize;
		Itm.PartialSize = I->CurrentSize;
		if (Itm.FileSize != 0)
			Itm.Percent = double (I->CurrentSize) / double (I->TotalSize);
		else
			Itm.Percent = 0;

		MessageSend(signalfd, MsgId_Update_Item, &Itm, sizeof(Itm));
	}

	struct MsgDownloadTotal DT;

	DT.CurrentBytes = CurrentBytes;
	DT.TotalBytes = TotalBytes;
	DT.CurrentItems = CurrentItems;
	DT.TotalItems = TotalItems;
	DT.CurrentCPS = CurrentCPS;
	DT.ETA = int((TotalBytes - CurrentBytes) / CurrentCPS);

	if ((TotalBytes + TotalItems) != 0)
		DT.Percent = double(CurrentBytes+CurrentItems) / double(TotalBytes+TotalItems);
	else
		DT.Percent = 0;

	MessageSend(signalfd, MsgId_Update_Total, &DT, sizeof(DT));
	MessageSend(signalfd, MsgId_Update, &J, sizeof(J));

	Update = false;

	return true;
}

void pkgAcquireBackend::Done(pkgAcquire::ItemDesc & Itm)
{
	struct MsgItemDesc Data;

	Data.FileSize = Itm.Owner->FileSize;
	Data.ID = Itm.Owner->ID;
	strncpy(Data.URI,         Itm.URI.c_str(), 200);
	strncpy(Data.Description, Itm.Description.c_str(), 200);
	strncpy(Data.ShortDesc,   Itm.ShortDesc.c_str(), 200);

	MessageSend(signalfd, MsgId_Download_Complete, &Data, sizeof(Data));

	Update = true;
}

bool pkgAcquireBackend::MediaChange(string Media, string Drive)
{
	struct MsgMediaChange NM;

	strncpy(NM.Title, Media.c_str(), 100);
	strncpy(NM.Drive, Drive.c_str(), 100);

	MessageSend(signalfd, MsgId_MediaChange, &NM, sizeof(NM));

	unsigned long len = sizeof(bool);
	bool result;
	MsgId Id;

	MessageReceive(signalfd, Id, &result, len);

	Update = true;

	return result;
}
